from typing import TypeVar

from sqlalchemy import Select
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.compiler import SQLCompiler

_T = TypeVar("_T")


def compile_stmt(stmt: Select[tuple[_T]]) -> SQLCompiler:
    return stmt.compile(
        dialect=postgresql.dialect(),  # type:ignore[no-untyped-call]
        compile_kwargs={"literal_binds": True},
    )



from typing import Annotated, assert_never

import pytest
from sqlalchemy import select

from sqla_filter.base import BaseSorter
from sqla_filter.ordering import OrderingEnum, OrderingField, build_priorities
from sqla_filter.unset import UNSET, Unset
from tests.sqla_filter.common.models import Author, Book, User
from tests.sqla_filter.common.ordering import BookSorter
from tests.utils import compile_stmt


@pytest.mark.parametrize("nulls", ["first", "last", None])
def test_build_simple_stmt(nulls: str | None) -> None:
    field = Book.created_at
    stmt = select(Book)

    expected_stmt = select(Book)
    match nulls:
        case "first":
            sorter = BookSorter(created_at_nulls_first=OrderingEnum.desc)
            expected_stmt = expected_stmt.order_by(field.desc().nulls_first())
        case "last":
            sorter = BookSorter(created_at_nulls_last=OrderingEnum.desc)
            expected_stmt = expected_stmt.order_by(field.desc().nulls_last())
        case None:
            sorter = BookSorter(created_at=OrderingEnum.desc)
            expected_stmt = expected_stmt.order_by(field.desc())
        case _ as never:
            raise assert_never(never)  # type: ignore[arg-type]

    stmt = sorter.apply(stmt)

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_stmt_with_join() -> None:
    stmt = select(Book)
    sorter = BookSorter(
        author_user_first_name=OrderingEnum.desc,
        author_user_is_deleted=OrderingEnum.asc,
    )
    stmt = sorter.apply(
        stmt,
        fields_priority=build_priorities(
            BookSorter.author_user_is_deleted,
            BookSorter.author_user_last_name,
        ),
    )

    expected_stmt = (
        select(Book)
        .join(Book.authors)
        .join(Author.user)
        .order_by(
            User.is_deleted.asc(),
            User.first_name.desc(),
        )
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_stmt_with_single_join() -> None:
    stmt = select(Book)
    sorter = BookSorter(author_alias=OrderingEnum.desc)
    stmt = sorter.apply(stmt)

    expected_stmt = select(Book).join(Book.authors).order_by(Author.alias.desc())
    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_sorter_raises() -> None:
    with pytest.raises(TypeError):

        class Sorter(BaseSorter):  # pyright:ignore[reportUnusedClass]
            created_at: Annotated[int | Unset, OrderingField(Book.created_at)] = UNSET



import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    TypeVar,
)

from sqlalchemy import ColumnElement, Select
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.operators import OperatorType

from .relationship import RelationshipInfo
from .types_ import SelectClause

if TYPE_CHECKING:
    from .base import BaseFilter

_TFilter = TypeVar("_TFilter", bound="BaseFilter")


class OperatorProtocol(Protocol):  # pragma: no cover
    def __call__(
        self,
        field: InstrumentedAttribute[Any],
        value: Any,  # noqa: ANN401
    ) -> ColumnElement[bool]: ...


class ManualFilter(ABC, Generic[SelectClause, _TFilter]):
    _name: str

    @abstractmethod
    def apply(
        self,
        stmt: Select[SelectClause],
        *,
        value: Any,  # noqa: ANN401
        filter_: _TFilter,
    ) -> Select[SelectClause]:
        raise NotImplementedError

    @property
    def name(self) -> str:  # pragma: no cover
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value


@dataclasses.dataclass(slots=True)
class FilterField:
    _name: str = dataclasses.field(init=False, repr=False, hash=True)

    field: InstrumentedAttribute[Any]
    _: dataclasses.KW_ONLY
    operator: OperatorType | OperatorProtocol
    relationship: RelationshipInfo | None = None
    """For filter by relationship field"""

    relationships: Sequence[RelationshipInfo] | None = None
    """For filter by nested relationship field"""

    @property
    def name(self) -> str:  # pragma: no cover
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value



import dataclasses
from typing import Any, TypeAlias

from sqlalchemy import ColumnExpressionArgument
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.roles import OnClauseRole

_OnClauseArgument: TypeAlias = ColumnExpressionArgument[Any] | OnClauseRole


@dataclasses.dataclass(frozen=True, slots=True)
class RelationshipInfo:
    field: InstrumentedAttribute[Any]
    _: dataclasses.KW_ONLY
    onclause: _OnClauseArgument | None = None
    isouter: bool = False
    full: bool = False



import enum
from typing import TypeVar

_T = TypeVar("_T")


class Unset(enum.Enum):
    v = enum.auto()


UNSET = Unset.v


def or_unset(value: _T | None) -> _T | Unset:
    if value is None:
        return UNSET

    return value


def define(value: _T | Unset) -> _T:
    if isinstance(value, Unset):
        msg = 'Got "Unset" type'
        raise TypeError(msg)

    return value



from .base import BaseFilter, BaseSorter, SupportsOrFilter
from .filter_ import (
    FilterField,
    ManualFilter,
    RelationshipInfo,
)
from .ordering import OrderingEnum, OrderingField
from .unset import UNSET, Unset, or_unset

__all__ = [
    "UNSET",
    "BaseFilter",
    "BaseSorter",
    "FilterField",
    "ManualFilter",
    "OrderingEnum",
    "OrderingField",
    "RelationshipInfo",
    "SupportsOrFilter",
    "Unset",
    "or_unset",
]



from datetime import datetime
from typing import TYPE_CHECKING, final

from sqlalchemy import Select

from sqla_filter.filter_ import ManualFilter
from tests.sqla_filter.common.models import Book

if TYPE_CHECKING:
    from examples.filter.example import BookFilter


@final
class BookManualFilter(ManualFilter[tuple[Book], "BookFilter"]):
    def apply(
        self,
        stmt: Select[tuple[Book]],
        *,
        value: bool,
        filter_: "BookFilter",  # noqa: ARG002
    ) -> Select[tuple[Book]]:
        if not value:
            return stmt

        return stmt.where(Book.created_at != datetime.min)  # noqa: DTZ901



import uuid
from datetime import UTC, datetime, timedelta

from sqlalchemy import or_, select

from tests.sqla_filter.common.filter import BookFilter, BookOrFilter, DateTimeInterval
from tests.sqla_filter.common.models import Author, Book, Review, User
from tests.utils import compile_stmt


def test_build_simple_stmt() -> None:
    ident = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookFilter(ident=ident)
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(Book.id == ident)

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_empty_or_stmt() -> None:
    ident = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookOrFilter(ident=ident, or_=BookOrFilter())
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(Book.id == ident)

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_empty_stmt() -> None:
    stmt = select(Book)
    filter_ = BookOrFilter()
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book)

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_simple_or_stmt() -> None:
    ident_1 = uuid.uuid4()
    ident_2 = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookOrFilter(ident=ident_1, or_=BookOrFilter(ident=ident_2))
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(
        or_(
            Book.id == ident_1,
            Book.id == ident_2,
        ),
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_nested_or_stmt() -> None:
    ident_1 = uuid.uuid4()
    ident_2 = uuid.uuid4()
    ident_3 = uuid.uuid4()
    ident_4 = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookOrFilter(
        ident=ident_1,
        or_=BookOrFilter(
            ident=ident_2,
            or_=BookOrFilter(
                ident=ident_3,
                or_=BookOrFilter(ident=ident_4),
            ),
        ),
    )
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(
        or_(
            Book.id == ident_1,
            Book.id == ident_2,
            Book.id == ident_3,
            Book.id == ident_4,
        ),
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_stmt_with_lambda() -> None:
    now = datetime.now(tz=UTC)
    tomorrow = now + timedelta(days=1)

    stmt = select(Book)
    filter_ = BookFilter(created_at_between=DateTimeInterval(from_=now, to=tomorrow))
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(Book.created_at.between(now, tomorrow))

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_stmt_with_join() -> None:
    ident = uuid.uuid4()
    author_ids = [uuid.uuid4() for _ in range(3)]
    review_ids = [uuid.uuid4() for _ in range(3)]

    stmt = select(Book)
    filter_ = BookFilter(ident=ident, author_ids=author_ids, review_ids=review_ids)
    stmt = filter_.apply(stmt)

    expected_stmt = (
        select(Book)
        .where(
            Book.id == ident,
            Author.id.in_(author_ids),
            Review.id.in_(review_ids),
        )
        .join(Book.authors)
        .join(Book.reviews)
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_stmt_with_multiple_join() -> None:
    user_id = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookFilter(author_user_id=user_id)
    stmt = filter_.apply(stmt)

    expected_stmt = (
        select(Book).where(User.id == user_id).join(Book.authors).join(Author.user)
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_manual_filter() -> None:
    ident = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookFilter(ident=ident, is_manual_filter_enabled=True)
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(
        Book.id == ident,
        Book.created_at != datetime.min,  # noqa: DTZ901
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string



import dataclasses
import enum
from collections.abc import Iterable, Sequence
from typing import Any, Literal, cast

from sqlalchemy.orm import InstrumentedAttribute

from .relationship import RelationshipInfo


class OrderingEnum(enum.StrEnum):
    asc = enum.auto()
    desc = enum.auto()


@dataclasses.dataclass(slots=True)
class OrderingField:
    _name: str = dataclasses.field(init=False, repr=False, hash=True)

    field: InstrumentedAttribute[Any]
    _: dataclasses.KW_ONLY
    relationship: RelationshipInfo | None = None
    """For ordering by relationship field"""

    relationships: Sequence[RelationshipInfo] | None = None
    """For ordering by nested relationship field"""

    nulls: Literal["first", "last"] | None = None

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value


def build_priorities(*fields: Any) -> Iterable[OrderingField]:  # noqa: ANN401
    """
    Hack for pyright

    Pass only `OrderingField`
    """
    return cast("Iterable[OrderingField]", fields)
